#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include "err.h"
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <strings.h>
#ifdef _WIN32
  #include <io.h>
  #include <fcntl.h>
#endif

typedef uint8_t u8;

int readbits = 4;
int copybits = 4;
int offsetbits = 4;
int countbits = 11;
int verbose = 0;

#define MINCOPY 2
#define MINOFFS 1

u8 unpacked[32768];
u8 packed[32768 * 8];
int packedpos;
int itemcount;
int ulength;
int readcount;
int readstart;
int pack_to_buf;

u8 bits;
int bitcount = 0;

#define SCR_WAITBITS	10
#define SCR_CMDBITS	3
#define SCR_FILEBITS	5

#define SCR_WAIT	0
#define SCR_SEEK	1
#define SCR_UNP_FILE	2
#define SCR_UNP_SCRIPT	3
#define SCR_LAUNCH	4
#define SCR_STOP	5
#define SCR_RESET	6
#define SCR_CLEAR	7

#define SCR_DEST_GFX	0
#define SCR_DEST_COG	1
#define SCR_DEST_EXTRA	2

struct bitfile {
	char	name[64];
	int	num;
} bitfile[2 << SCR_FILEBITS];

int nbitfiles = 0;

char *filename = "";

void putbyte(u8 v) {
	fputc(v, stdout);
}

void putbit(int v) {
	if(pack_to_buf) {
		packed[packedpos++] = !!v;
	} else {
		bits >>= 1;
		if(v) bits |= 0x80;
		bitcount++;
		if(bitcount == 8) {
			putbyte(bits);
			bitcount = 0;
		}
	}
}

void chunk(unsigned int v, int width) {
	int i;

	if(v >= (1 << width)) errx(1, "(%s) %d doesn't fit in %d bits", filename, v, width);
	for(i = width - 1; i >= 0; i--) {
		putbit(v & (1 << i));
	}
}

int matchlength(int a, int b, int max) {
	int i;

	if(max > (1 << copybits) + MINCOPY - 1) max = (1 << copybits) + MINCOPY - 1;

	for(i = 0; i < max; i++) {
		if(unpacked[a + i] != unpacked[b + i]) return i;
	}
	return max;
}

void flushread() {
	int i;

	if(readcount) {
		chunk(0, 1);
		chunk(readcount - 1, readbits);
		for(i = 0; i < readcount; i++) {
			chunk(unpacked[readstart++], 8);
		}
		//fprintf(stderr, "READ %d\n", readcount);
		readcount = 0;
		itemcount++;
	}
}

void packfile() {
	int i, offset;
	int bestoffset, bestlength;
	int m;

	itemcount = 0;
	packedpos = 0;
	readcount = 0;
	readstart = 0;

	pack_to_buf = 1;
	for(i = 0; i < ulength;) {
		bestoffset = 0;
		bestlength = 0;
		for(offset = MINOFFS; offset < ((1 << offsetbits) + MINOFFS); offset++) {
			if(i - offset >= 0) {
				m = matchlength(i, i - offset, ulength - i);
				if(m > bestlength) {
					bestlength = m;
					bestoffset = offset;
				}
			}
		}
		if(bestlength > MINCOPY) {
			flushread();
			//fprintf(stderr, "COPY %d from %d\n", bestlength, bestoffset);
			chunk(1, 1);
			chunk(bestoffset - MINOFFS, offsetbits);
			chunk(bestlength - MINCOPY, copybits);
			itemcount++;
			readstart += bestlength;
			i += bestlength;
		} else {
			readcount++;
			if(readcount == (1 << readbits)) flushread();
			i++;
		}
	}
	flushread();

	pack_to_buf = 0;
	chunk(itemcount, countbits);
	for(i = 0; i < packedpos; i++) {
		putbit(packed[i]);
	}
}

uint8_t parsecolor(int x) {
	return ((x / 100) << 4) | (((x / 10) % 10) << 2) | (x % 10);
}

void parsescript(FILE *fp) {
	char buf[256], s1[256];
	int lastclock = 0, clock;
	int p1, p2, p3, p4;
	int i;

	pack_to_buf = 0;

	while(!feof(fp) && fgets(buf, sizeof(buf), fp)) {
		again:
		if(2 == sscanf(buf, "%x:%x", &p1, &p2)) {
			clock = p1 * 32 + p2;
			if(clock > lastclock) {
				chunk(SCR_WAIT, SCR_CMDBITS);
				chunk(clock - lastclock, SCR_WAITBITS);
				lastclock = clock;
			}
		} else if(2 == sscanf(buf, " launch %d, %d", &p1, &p2)) {
			chunk(SCR_LAUNCH, SCR_CMDBITS);
			chunk(p1, 3);
			chunk(p2, 1);
		} else if(1 == sscanf(buf, " stop %d", &p1)) {
			chunk(SCR_STOP, SCR_CMDBITS);
			chunk(p1, 3);
		} else if(1 == sscanf(buf, " %s", s1) && !strcmp(s1, "clear")) {
			chunk(SCR_CLEAR, SCR_CMDBITS);
		} else if(1 == sscanf(buf, " %s", s1) && !strcmp(s1, "reset")) {
			chunk(SCR_RESET, SCR_CMDBITS);
		} else if(1 == sscanf(buf, " unpack %s", s1) && !strcmp(s1, "cog")) {
			chunk(SCR_UNP_FILE, SCR_CMDBITS);
			chunk(SCR_DEST_COG, 2);
		} else if(1 == sscanf(buf, " unpack %s", s1) && !strcmp(s1, "extra")) {
			chunk(SCR_UNP_FILE, SCR_CMDBITS);
			chunk(SCR_DEST_EXTRA, 2);
		} else if(1 == sscanf(buf, " unpack %s", s1) && !strcmp(s1, "gfx")) {
			chunk(SCR_UNP_FILE, SCR_CMDBITS);
			chunk(SCR_DEST_GFX, 2);
		} else if(1 == sscanf(buf, " seek %s", s1)) {
			chunk(SCR_SEEK, SCR_CMDBITS);
			for(i = 0; i < nbitfiles; i++) {
				if(!strcasecmp(s1, bitfile[i].name)) {
					break;
				}
			}
			if(i == nbitfiles) errx(1, "Unknown bitfile: \"%s\"", s1);
			chunk(bitfile[i].num, SCR_FILEBITS);
		} else if(1 == sscanf(buf, " %s", s1) && !strcmp(s1, "palette")) {
			int gotline = 0;

			chunk(SCR_UNP_SCRIPT, SCR_CMDBITS);
			ulength = 0;
			while(!feof(fp) && fgets(buf, sizeof(buf), fp)) {
				if(4 == sscanf(buf, " %X, %X, %X, %X", &p1, &p2, &p3, &p4)) {
					unpacked[ulength++] = p1^0xF0;//parsecolor(p1);
					unpacked[ulength++] = p2^0xF0;//parsecolor(p2);
					unpacked[ulength++] = p3^0xF0;//parsecolor(p3);
					unpacked[ulength++] = p4^0xF0;//parsecolor(p4);
				} else {
					gotline = 1;
					break;
				}
			}
			packfile();
			if(gotline) goto again;		// here there be raptors
		}
	}
	chunk(SCR_WAIT, SCR_CMDBITS);
	chunk(0, SCR_WAITBITS);
}

int main(int argc, char **argv) {
	int opt, arg;
	int script = 0;
#ifdef _WIN32
	_setmode(fileno(stdout), O_BINARY);
#endif
	while((opt = getopt(argc, argv, "r:c:o:n:s:v")) >= 0) {
		switch(opt) {
			case 'r':
				readbits = atoi(optarg);
				break;
			case 'c':
				copybits = atoi(optarg);
				break;
			case 'o':
				offsetbits = atoi(optarg);
				break;
			case 'n':
				countbits = atoi(optarg);
				break;
			case 's':
				script = 1;
				{
					char buf[256];
					FILE *fp = fopen(optarg, "r");
					if(!fp) err(1, optarg);
					while(!feof(fp) && fgets(buf, sizeof(buf), fp)) {
						if(2 == sscanf(buf, "BF_%s = %d", bitfile[nbitfiles].name, &bitfile[nbitfiles].num)) {
							nbitfiles++;
						}
					}
					fclose(fp);
				}
				break;
			case 'v':
				verbose = 1;
				break;
			default:
				errx(1, "Usage: pack -r readbits -c copybits -o offsetbits -n countbits [-s demo.i] [-v]");
		}
	}

	for(arg = optind; arg < argc; arg++) {
		FILE *fp = fopen(argv[arg], "rb");
		if(!fp) err(1, "%s", argv[arg]);
		filename = argv[arg];
		if(script) {
			parsescript(fp);
		} else {
			ulength = fread(unpacked, 1, sizeof(unpacked), fp);
			packfile();
		}
		fclose(fp);
		if(verbose && !script) {
			int newsize = ((packedpos + countbits) + 7) / 8;
			fprintf(stderr, "%-22s Original size: %5d  Compressed size: %5d  Ratio: %6.2f%%  Item count: %d\n",
				argv[arg],
				ulength,
				newsize,
				newsize * 100.0 / ulength,
				itemcount);
		}
	}

	while(bitcount) putbit(0);

	return 0;
}
